import os

import fbnconfig

if os.getenv("FBN_ACCESS_TOKEN") is None or os.getenv("LUSID_ENV") is None:
    raise (
        RuntimeError("Both FBN_ACCESS_TOKEN and LUSID_ENV variables need to be set to run these tests")
    )

lusid_env = os.environ["LUSID_ENV"]
token = os.environ["FBN_ACCESS_TOKEN"]


def test_setup():
    client = fbnconfig.create_client(lusid_env, token)
    fbnconfig.setup(client)
    # check that something is returned
    assert len(fbnconfig.list_deployments(client)) >= 0



import os
import pathlib
from types import SimpleNamespace

import pytest
from pytest import fixture

import fbnconfig
from fbnconfig import drive
from tests.integration.generate_test_name import gen


@fixture
def lusid_env():
    if os.getenv("FBN_ACCESS_TOKEN") is None or os.getenv("LUSID_ENV") is None:
        raise (
            RuntimeError("FBN_ACCESS_TOKEN and LUSID_ENV variables need to be set to run these tests")
        )

    env = os.environ["LUSID_ENV"]
    token = os.environ["FBN_ACCESS_TOKEN"]
    return SimpleNamespace(env=env, token=token)


@fixture
def client(lusid_env):
    return fbnconfig.create_client(lusid_env.env, lusid_env.token)


@fixture
def deployment_name():
    return gen("drive")


def resources(deployment_name):
    base_folder = deployment_name
    f1 = drive.FolderResource(id="base_folder", name=base_folder, parent=drive.root)
    f2 = drive.FolderResource(id="sub_folder", name="subfolder", parent=f1)
    f3 = drive.FolderResource(id="sub_sub_folder", name="subfolder2", parent=f2)
    content_path = pathlib.Path(__file__).parent.resolve() / pathlib.Path("poem.txt")
    ff = drive.FileResource(id="file1", folder=f3, name="myfile.txt", content_path=content_path)
    return {r.id: r for r in [f1, f2, f3, ff]}


@fixture
def deployment(deployment_name, lusid_env):
    res = resources(deployment_name)
    print(f"\nRunning for deployment {deployment_name}...")
    yield fbnconfig.Deployment(deployment_name, list(res.values()))
    print("\nTearing down resources...")
    fbnconfig.deployex(fbnconfig.Deployment(deployment_name, []), lusid_env.env, lusid_env.token)


def test_teardown(deployment, client, lusid_env):
    fbnconfig.deployex(deployment, lusid_env.env, lusid_env.token)
    fbnconfig.deployex(fbnconfig.Deployment(deployment.id, []), lusid_env.env, lusid_env.token)
    search = client.post("/drive/api/search/", json={"withPath": "/", "name": deployment.id})
    assert search.json()["values"] == []


def test_create(deployment, client, lusid_env):
    fbnconfig.deployex(deployment, lusid_env.env, lusid_env.token)
    search = client.post("/drive/api/search/", json={"withPath": "/", "name": deployment.id})
    assert len(search.json()["values"]) == 1


def test_update(deployment, lusid_env, client):
    fbnconfig.deployex(deployment, lusid_env.env, lusid_env.token)
    search = client.post("/drive/api/search/", json={"withPath": "/", "name": deployment.id})
    assert len(search.json()["values"]) == 1


@pytest.mark.asyncio
async def test_file_fetch(deployment, lusid_env):
    # GIVEN File is deployed
    fbnconfig.deployex(deployment, lusid_env.env, lusid_env.token)
    deployment_name = deployment.id
    client = fbnconfig.create_client(lusid_env.env, lusid_env.token, is_async=True)
    search = await client.post(
        "/drive/api/search/",
        json={"withPath": f"/{deployment_name}/subfolder/subfolder2", "name": "myfile.txt"}
    )
    drive_id = search.json()["values"][0]["id"]
    # WHEN the fetch method is called for that resource
    response = await drive.FileResource.fetch(
        client=client,
        keys={
            "driveId": drive_id,
        }
    )
    # THEN the response is as expected
    expected_response = {
        "id": f"file_{drive_id}".lower(),
        "driveId": drive_id,
        "folder": {
            "$ref": f"folder_/{deployment_name}/subfolder/subfolder2"
        },
        "name": "myfile.txt",
        "contentPath": f"/{deployment_name}/subfolder/subfolder2/myfile.txt",
    }
    assert response == expected_response


@pytest.mark.asyncio
async def test_folder_list_root_folder(deployment, lusid_env):
    # GIVEN Folder is deployed
    fbnconfig.deployex(deployment, lusid_env.env, lusid_env.token)
    deployment_name = deployment.id
    client = fbnconfig.create_client(lusid_env.env, lusid_env.token, is_async=True)
    search = await client.post(
        "/drive/api/search/",
        json={"withPath": "/", "name": deployment_name}
    )
    drive_id = search.json()["values"][0]["id"]
    # WHEN the list method is called for that resource
    response = await drive.FolderResource.list(
        client=client,
        query_params={"filter": f"name eq '{deployment_name}'"},
        path_params={}
    )
    # THEN the response is as expected
    expected_response = {
        "values": [
            {
                "keys": {
                    "driveId": drive_id,
                },
                "value": {
                    "id": f"folder_/{deployment_name}",
                    "driveId": drive_id,
                    "name": deployment_name,
                    "parent": {
                        "$ref": drive.root.id
                    }
                }
            }
        ]
    }
    assert response == expected_response


@pytest.mark.asyncio
async def test_folder_list_not_root_folder(deployment, lusid_env):
    # GIVEN Folder is deployed
    fbnconfig.deployex(deployment, lusid_env.env, lusid_env.token)
    deployment_name = deployment.id
    client = fbnconfig.create_client(lusid_env.env, lusid_env.token, is_async=True)
    search = await client.post(
        "/drive/api/search/",
        json={"withPath": "/", "name": deployment_name}
    )
    parent_drive_id = search.json()["values"][0]["id"]
    # WHEN the list method is called for that resource
    response = await drive.FolderResource.list(
        client=client,
        query_params={"filter": "name eq 'subfolder'"},
        path_params={
            "driveId": parent_drive_id
        }
    )
    # THEN the response is as expected
    search = await client.post(
        "/drive/api/search/",
        json={"withPath": f"/{deployment_name}", "name": "subfolder"}
    )
    drive_id = search.json()["values"][0]["id"]
    expected_response = {
        "values": [
            {
                "keys": {
                    "driveId": drive_id,
                },
                "value": {
                    "id": f"folder_/{deployment_name}/subfolder",
                    "driveId": drive_id,
                    "name": "subfolder",
                    "parent": {
                        "$ref": f"folder_/{deployment_name}"
                    }
                }
            }
        ]
    }
    assert response == expected_response


@pytest.mark.asyncio
async def test_folder_fetch(deployment, lusid_env):
    # GIVEN Folder is deployed
    fbnconfig.deployex(deployment, lusid_env.env, lusid_env.token)
    deployment_name = deployment.id
    client = fbnconfig.create_client(lusid_env.env, lusid_env.token, is_async=True)
    search = await client.post(
        "/drive/api/search/",
        json={"withPath": f"/{deployment_name}/subfolder", "name": "subfolder2"}
    )
    drive_id = search.json()["values"][0]["id"]
    # WHEN the fetch method is called for that resource
    response = await drive.FolderResource.fetch(
        client=client,
        keys={
            "driveId": drive_id,
        }
    )
    # THEN the response is as expected
    expected_response = {
        "id": f"folder_/{deployment_name}/subfolder/subfolder2",
        "driveId": drive_id,
        "name": "subfolder2",
        "parent": {
            "$ref": f"folder_/{deployment_name}/subfolder"
        },
    }
    assert response == expected_response



from __future__ import annotations

import copy
import json
from enum import StrEnum
from hashlib import sha256
from typing import Annotated, Any, ClassVar, Dict, List

import httpx
from pydantic import (
    BaseModel,
    BeforeValidator,
    Field,
    PlainSerializer,
    WithJsonSchema,
    model_serializer,
    model_validator,
)

from . import property  # type: noqa
from .properties import MetricValue, PropertyValue, PropertyValueDict
from .resource_abc import CamelAlias, FetchMixin, ListMixin, Ref, Resource, register_resource
from .urlfmt import Urlfmt

# force these to be used so they appear in the customentity namespace
_ = [property, PropertyValue, MetricValue]

# Context key for reference resolution
REFS_CONTEXT_KEY = "$refs"
GROUP = "Data Model"


class CollectionType(StrEnum):
    SINGLE = "Single"
    ARRAY = "Array"


class LifeTime(StrEnum):
    PERPETUAL = "Perpetual"
    TIMEVARIANT = "TimeVariant"


class FieldType(StrEnum):
    STRING = "String"
    BOOLEAN = "Boolean"
    DATE_TIME = "DateTime"
    DECIMAL = "Decimal"


class FieldDefinition(CamelAlias, BaseModel):
    name: str
    lifetime: LifeTime
    type: FieldType
    collection_type: CollectionType = CollectionType.SINGLE
    required: bool
    description: str = ""


# These are optional in the API create and will be given default values. When read is called
# they will not be returned if they have the default value
DEFAULT_FIELD = {"collectionType": "Single", "description": ""}


@register_resource(group=GROUP, keys=["entityType"])
class EntityTypeResource(CamelAlias, BaseModel, Resource, ListMixin, FetchMixin):
    id: str = Field(exclude=True)
    entity_type_name: str
    display_name: str
    description: str
    field_schema: List[FieldDefinition]

    @model_validator(mode="before")
    @classmethod
    def des_model(cls, data: Any, info) -> Dict[str, Any]:
        if not isinstance(data, dict):
            return data
        if info.context and info.context.get("id"):
            return data | {"id": info.context.get("id")}
        return data

    def read(self, client, old_state) -> Dict[str, Any]:
        entity_type = old_state.entitytype
        return client.request("get", f"/api/api/customentities/entitytypes/{entity_type}").json()

    def create(self, client: httpx.Client):
        desired = self.model_dump(mode="json", exclude_none=True, by_alias=True)
        res = client.request("POST", "/api/api/customentities/entitytypes", json=desired).json()
        return {"entitytype": res["entityType"]}

    def update(self, client: httpx.Client, old_state):
        remote = self.read(client, old_state)
        # enrich remote fields with the default values if not present
        remote["fieldSchema"] = [rem | DEFAULT_FIELD for rem in remote["fieldSchema"]]
        desired = self.model_dump(mode="json", exclude_none=True, by_alias=True)
        effective = remote | copy.deepcopy(desired)
        for i in range(0, len(self.field_schema)):
            if i < len(remote["fieldSchema"]):
                eff_field = remote["fieldSchema"][i] | desired["fieldSchema"][i]
                effective["fieldSchema"][i] = eff_field
        if effective == remote:
            return None
        res = client.request(
            "PUT", f"/api/api/customentities/entitytypes/{old_state.entitytype}", json=desired
        ).json()
        return {"entitytype": res["entityType"]}

    @staticmethod
    def delete(client, old_state):
        raise RuntimeError("Cannot delete a custom entity definition")

    @classmethod
    def create_resource_id(cls, data: dict) -> str:
        return f"entity-type_{data['entityType']}".lower()

    @classmethod
    def get_keys(cls, data: dict[str, Any]) -> dict[str, Any]:
        return {"entityType": data["entityType"]}

    @classmethod
    def from_api(cls, data: dict[str, Any]) -> dict[str, Any]:
        result = {
            "id": cls.create_resource_id(data),
            "entityTypeName": data["entityTypeName"],
            "displayName": data["displayName"],
            "description": data["description"],
            "fieldSchema": []
        }

        # Process fieldSchema, removing default values
        for field in data.get("fieldSchema", []):
            field_data = {
                "name": field["name"],
                "lifetime": field["lifetime"],
                "type": field["type"],
                "required": field["required"]
            }
            # Only include non-default values
            if field.get("collectionType") and field.get("collectionType") != "Single":
                field_data["collectionType"] = field["collectionType"]
            if field.get("description", "") != "":
                field_data["description"] = field["description"]

            result["fieldSchema"].append(field_data)

        return result

    @classmethod
    def fetch_endpoint(cls) -> str:
        return "/api/api/customentitytypes/{entityType}"

    @classmethod
    def list_endpoint(cls, **kwargs) -> str:
        return "/api/api/customentitytypes"

    def deps(self):
        return []


@register_resource(group=GROUP)
class EntityTypeRef(CamelAlias, BaseModel, Ref):
    id: str = Field(exclude=True)
    entity_type_name: str

    def attach(self, client):
        entity_type = self.entity_type_name
        try:
            client.get(f"/api/api/customentities/entitytypes/{entity_type}").json()
        except httpx.HTTPStatusError as ex:
            if ex.response.status_code == 404:
                raise RuntimeError(f"Custom Entity Definition {entity_type} does not exist")
            else:
                raise ex


def ser_entitytype_key(value, info):
    if info.context and info.context.get("style", "api") == "dump":
        return {"$ref": value.id}
    return "~" + value.entity_type_name


def des_entitytype_key(value, info):
    if not isinstance(value, dict):
        return value
    if info.context and info.context.get(REFS_CONTEXT_KEY):
        ref = info.context[REFS_CONTEXT_KEY][value["$ref"]]
        return ref
    return value


EntityTypeKey = Annotated[
    EntityTypeResource | EntityTypeRef,
    BeforeValidator(des_entitytype_key),
    PlainSerializer(ser_entitytype_key),
    WithJsonSchema(
        {
            "type": "object",
            "properties": {"$ref": {"type": "string", "format": "Key.EntityType"}},
            "required": ["$ref"],
        }
    ),
]


class EntityField(CamelAlias, BaseModel):
    effective_from: str | None = None
    effective_to: str | None = None
    name: str
    value: Any


class EntityIdentifier(CamelAlias, BaseModel):
    effective_from: str | None = None
    effective_to: str | None = None
    identifier_type: property.DefinitionRef | property.DefinitionResource
    identifier_value: str

    @model_serializer()
    def ser_model(self, info):
        style = info.context.get("style", "api") if info.context else "api"
        if style == "dump":
            return {
                "identifierType": {"$ref": self.identifier_type.id},
                "identifierValue": self.identifier_value,
            }
        return {
            "identifierScope": self.identifier_type.scope,
            "identifierType": self.identifier_type.code,
            "identifierValue": self.identifier_value
        }

    @model_validator(mode="before")
    @classmethod
    def des_model(cls, data, info):
        if not isinstance(data, dict):
            return data
        if info.context and info.context.get(REFS_CONTEXT_KEY):
            ident_type = data["identifierType"]
            if isinstance(ident_type, dict) and "$ref" in ident_type:
                ref_id = ident_type["$ref"]
                return data | {"identifierType": info.context[REFS_CONTEXT_KEY][ref_id]}
            elif isinstance(ident_type, str):
                return data | {"identifierType": info.context[REFS_CONTEXT_KEY][ident_type]}
        return data


@register_resource(
    group=GROUP,
    parent=EntityTypeResource,
    keys=["entityType", "identifierType", "identifierValue", "identifierScope"]
)
class EntityResource(CamelAlias, BaseModel, Resource, ListMixin, FetchMixin):
    id: str = Field(exclude=True)
    fields: List[EntityField]
    identifiers: List[EntityIdentifier] = Field(min_length=1)
    entity_type: EntityTypeKey
    description: str
    display_name: str
    properties: PropertyValueDict | None = None

    u: ClassVar = Urlfmt("/api/api/customentities")

    @model_validator(mode="before")
    @classmethod
    def des_model(cls, data, info):
        if not isinstance(data, dict):
            return data
        if info.context and info.context.get("id"):
            data = data | {"id": info.context.get("id")}
        return data

    def create(self, client):
        desired = self.model_dump(mode="json", exclude_none=True, by_alias=True, exclude={"entity_type"})
        sorted_desired = json.dumps(desired, sort_keys=True)
        content_hash = sha256(sorted_desired.encode()).hexdigest()
        i0 = desired["identifiers"][0]
        url = self.u.format("{base}/~{type}", type=self.entity_type.entity_type_name)
        client.post(url, json=desired)
        return {
            "entity_type": self.entity_type.entity_type_name,
            "identifier_scope": i0["identifierScope"],
            "identifier_type": i0["identifierType"],
            "identifier_value": i0["identifierValue"],
            "content_hash": content_hash
        }

    def read(self, client, old_state) -> Dict:
        url = self.u.format("{base}/~{type}/{idtype}/{idvalue}",
            type=old_state.entity_type,
            idtype=old_state.identifier_type,
            idvalue=old_state.identifier_value,
        )
        res = client.get(url, params={"identifierScope": old_state.identifier_scope}).json()
        res.pop("href", None)
        res.pop("version", None)
        res.pop("relationships", None)
        return res

    def update(self, client, old_state):
        desired = self.model_dump(mode="json", exclude_none=True, by_alias=True,
                                  exclude={"id", "scope", "entity_type"})
        sorted_desired = json.dumps(desired, sort_keys=True)
        content_hash = sha256(sorted_desired.encode()).hexdigest()
        if old_state.content_hash == content_hash:
            return None
        # if the type or previously used identifier has changes delete and recreate
        newid = [
            i for i in desired["identifiers"]
             if i["identifierType"] == old_state.identifier_type and
                i["identifierScope"] == old_state.identifier_scope and
                i["identifierValue"] == old_state.identifier_value
        ]
        if len(newid) != 1 or self.entity_type.entity_type_name != old_state.entity_type:
            self.delete(client, old_state)
            return self.create(client)
        url = self.u.format("{base}/~{type}", type=self.entity_type.entity_type_name)
        client.post(url, json=desired)
        i0 = desired["identifiers"][0]
        return {
            "entity_type": self.entity_type.entity_type_name,
            "identifier_scope": i0["identifierScope"],
            "identifier_type": i0["identifierType"],
            "identifier_value": i0["identifierValue"],
            "content_hash": content_hash
        }

    @staticmethod
    def delete(client, old_state):
        url = EntityResource.u.format("{base}/~{type}/{idtype}/{idvalue}",
            type=old_state.entity_type,
            idtype=old_state.identifier_type,
            idvalue=old_state.identifier_value,
        )
        client.delete(url, params={"identifierScope": old_state.identifier_scope})

    def deps(self):
        res = [self.entity_type] + [i.identifier_type for i in self.identifiers]
        if self.properties:
            res.extend([p.property_key for p in self.properties])
        return res

    @classmethod
    def list_endpoint(cls, **kwargs) -> str:
        return "/api/api/customentities/{entityType}"

    @classmethod
    def fetch_endpoint(cls) -> str:
        return "/api/api/customentities/{entityType}/{identifierType}/{identifierValue}"

    @classmethod
    def create_resource_id(cls, data: dict) -> str:
        identifier = data["identifiers"][0]
        id_parts = "_".join(
            identifier[key] for key in ("identifierType", "identifierValue", "identifierScope")
        )
        return f"entity_{data['entityType']}_{id_parts}".lower()

    @classmethod
    def get_keys(cls, data: dict[str, Any]) -> dict[str, Any]:
        identifier = data["identifiers"][0]
        return {
            "identifierType": identifier["identifierType"],
            "identifierValue": identifier["identifierValue"],
            "identifierScope": identifier["identifierScope"],
            "entityType": data["entityType"]
        }

    @classmethod
    def from_api(cls, data: dict[str, Any]) -> dict[str, Any]:
        data.pop("href", None)
        data.pop("links", None)
        data.pop("version", None)
        data.pop("relationships", None)
        data.pop("stagedModifications", None)
        data.pop("identifierScope", None)
        data.pop("identifierType", None)
        data.pop("identifierValue", None)

        data["id"] = cls.create_resource_id(data)

        entity_type = data.pop("entityType")
        data["entityType"] = {
            "$ref": f"entity-type_{entity_type}".lower()
        }

        identifier = data.pop("identifiers")[0]
        data["identifiers"] = [{
            "identifierType": {
                "$ref": f"identifier-type_{identifier['identifierType']}".lower()
            },
            "identifierValue": identifier["identifierValue"],
            "effectiveFrom": identifier["effectiveFrom"],
            "effectiveTo": identifier["effectiveUntil"]
        }]

        for field in data["fields"]:
            field["effectiveTo"] = field.pop("effectiveUntil")

        return data



import os
from types import SimpleNamespace

import pytest
from httpx import HTTPStatusError

import fbnconfig
import tests.integration.recipe as recipe
from tests.integration.generate_test_name import gen

if os.getenv("FBN_ACCESS_TOKEN") is None or os.getenv("LUSID_ENV") is None:
    raise (
        RuntimeError("Both FBN_ACCESS_TOKEN and LUSID_ENV variables need to be set to run these tests")
    )

lusid_env = os.environ["LUSID_ENV"]
token = os.environ["FBN_ACCESS_TOKEN"]
host_vars = {}
client = fbnconfig.create_client(lusid_env, token)


@pytest.fixture
def setup_deployment():
    deployment_name = gen("recipe")
    print(f"\nRunning for deployment {deployment_name}...")
    yield SimpleNamespace(name=deployment_name)
    # Teardown: Clean up resources (if any) after the test
    print("\nTearing down resources...")
    fbnconfig.deployex(fbnconfig.Deployment(deployment_name, []), lusid_env, token)


@pytest.fixture
def recipe_dict():
    return {
        "market": {
            "marketRules": [
                {
                    "key": "Fx.CurrencyPair.*",
                    "supplier": "DataScope",
                    "dataScope": "SomeScopeToLookAt",
                    "quoteType": "Rate",
                    "field": "Mid",
                    "priceSource": "",
                    "sourceSystem": "Lusid",
                    "fallThroughOnAccessDenied": False
                }
            ],
            "suppliers": {},
            "options": {
                "defaultSupplier": "Lusid",
                "defaultInstrumentCodeType": "LusidInstrumentId",
                "defaultScope": "default",
                "attemptToInferMissingFx": False,
                "calendarScope": "CoppClarkHolidayCalendars",
                "conventionScope": "Conventions"
            },
            "specificRules": [],
            "groupedMarketRules": []
        },
        "pricing": {
            "modelRules": [],
            "modelChoice": {},
            "options": {
                "modelSelection": {
                    "library": "Lusid",
                    "model": "SimpleStatic"
                },
                "useInstrumentTypeToDeterminePricer": False,
                "allowAnyInstrumentsWithSecUidToPriceOffLookup": False,
                "allowPartiallySuccessfulEvaluation": False,
                "produceSeparateResultForLinearOtcLegs": False,
                "enableUseOfCachedUnitResults": False,
                "windowValuationOnInstrumentStartEnd": False,
                "removeContingentCashflowsInPaymentDiary": False,
                "useChildSubHoldingKeysForPortfolioExpansion": False,
                "validateDomesticAndQuoteCurrenciesAreConsistent": False,
                "mbsValuationUsingHoldingCurrentFace": False,
                "convertSrsCashFlowsToPortfolioCurrency": False,
                "conservedQuantityForLookthroughExpansion": "PV",
                "returnZeroPv": {
                    "instrumentMatured": False
                },
                "enableLegLevelInferenceForCustomSrsColumns": False,
                "useInstrumentScaleFactorAsDefault": False
            },
            "resultDataRules": [],
            "holdingPricingInfo": {
                "fallbackField": "None",
                "overrideField": "None",
                "specificFallbacks": [],
                "specificOverrides": []
            }
        },
        "aggregation": {
            "options": {
                "useAnsiLikeSyntax": False,
                "allowPartialEntitlementSuccess": False,
                "applyIso4217Rounding": False
            }
        },
        "description": "",
        "holding": {
            "taxLotLevelHoldings": True
        }
    }


def test_teardown(setup_deployment):
    # create first
    deployment_name = setup_deployment.name
    fbnconfig.deployex(recipe.configure(setup_deployment), lusid_env, token)
    fbnconfig.deployex(fbnconfig.Deployment(deployment_name, []), lusid_env, token)
    with pytest.raises(HTTPStatusError) as error:
        client.get(f"/api/api/recipes/sc1/cd1_{deployment_name}")
    assert error.value.response.status_code == 404
    with pytest.raises(HTTPStatusError) as error:
        client.get(f"/api/api/recipes/sc1/cd2_{deployment_name}")
    assert error.value.response.status_code == 404


def test_create(setup_deployment):
    deployment_name = setup_deployment.name
    fbnconfig.deployex(recipe.configure(setup_deployment), lusid_env, token)
    search_1 = client.get(f"/api/api/recipes/sc1/cd1_{deployment_name}")
    assert search_1.status_code == 200
    search_2 = client.get(f"/api/api/recipes/sc1/cd2_{deployment_name}")
    assert search_2.status_code == 200


def test_update(setup_deployment):
    deployment_name = setup_deployment.name
    fbnconfig.deployex(recipe.configure(setup_deployment), lusid_env, token)
    update = fbnconfig.deployex(recipe.configure(setup_deployment), lusid_env, token)
    search_1 = client.get(f"/api/api/recipes/sc1/cd1_{deployment_name}")
    assert search_1.status_code == 200
    search_2 = client.get(f"/api/api/recipes/sc1/cd2_{deployment_name}")
    assert search_2.status_code == 200
    assert [a.change for a in update if a.type == "RecipeResource"] == ["nochange", "nochange"]


@pytest.mark.asyncio
async def test_recipe_list_with_filter(setup_deployment, recipe_dict):
    # GIVEN Recipe is deployed
    fbnconfig.deployex(recipe.configure(setup_deployment), lusid_env, token)
    scope = "sc1"
    code = f"cd1_{setup_deployment.name}"
    client = fbnconfig.create_client(lusid_env, token, is_async=True)
    # WHEN calling list with filter for the deployment name
    result = await recipe.recipe.RecipeResource.list(
        client=client,
        query_params={"filter": f"value.code eq '{code}'"},
        path_params={}
    )
    # THEN the result has correct structure
    expected_result = {
        "values": [
            {
                "keys": {
                    "scope": scope,
                    "code": code,
                },
                "value": {
                    "id": f"recipe_{scope}_{code}".lower(),
                    "scope": scope,
                    "code": code,
                    "recipe": recipe_dict
                }
            }
        ]
    }
    assert result["values"] == expected_result["values"]


@pytest.mark.asyncio
async def test_recipe_fetch(setup_deployment, recipe_dict):
    # GIVEN Recipe is deployed
    fbnconfig.deployex(recipe.configure(setup_deployment), lusid_env, token)
    scope = "sc1"
    code = f"cd1_{setup_deployment.name}"
    client = fbnconfig.create_client(lusid_env, token, is_async=True)
    # WHEN the fetch method is called for that resource
    response = await recipe.recipe.RecipeResource.fetch(
        client=client,
        keys={
            "scope": scope,
            "code": code
        }
    )
    # THEN the response is as expected
    expected_response = {
        "id": f"recipe_{scope}_{code}".lower(),
        "scope": scope,
        "code": code,
        "recipe": recipe_dict
    }
    assert response == expected_response



from .deploy import Deployment as Deployment
from .deploy import deploy as deploy
from .deploy import deployex as deployex
from .deploy import dump_deployment as dump_deployment
from .deploy import load_vars as load_vars
from .deploy import undump_deployment as undump_deployment
from .http_client import create_client as create_client
from .log import list_deployments as list_deployments
from .log import setup as setup


# import all resources so that the registration gets run.
from fbnconfig import access
from fbnconfig import cleardown_module
from fbnconfig import compliance
from fbnconfig import configuration
from fbnconfig import custodian_accounts
from fbnconfig import corporate_action
from fbnconfig import custom_data_model
from fbnconfig import customentity
from fbnconfig import cutlabel
from fbnconfig import datatype
from fbnconfig import drive
from fbnconfig import fund_accounting
from fbnconfig import fund_configurations
from fbnconfig import general_ledger_profile
from fbnconfig import horizon
from fbnconfig import identifier_definition
from fbnconfig import identity
from fbnconfig import lumi
from fbnconfig import notifications
from fbnconfig import posting_module
from fbnconfig import property
from fbnconfig import recipe
from fbnconfig import scheduler
from fbnconfig import sequence
from fbnconfig import side_definition
from fbnconfig import transaction_type
from fbnconfig import workflows
from fbnconfig import workspace
from fbnconfig import relational_dataset_definition
from fbnconfig import instrument_events



import os
from types import SimpleNamespace

import pytest
from httpx import HTTPStatusError
from pytest import fixture

import fbnconfig
import tests.integration.configuration as configuration
from tests.integration.generate_test_name import gen

if os.getenv("FBN_ACCESS_TOKEN") is None or os.getenv("LUSID_ENV") is None:
    raise (
        RuntimeError("Both FBN_ACCESS_TOKEN and LUSID_ENV variables need to be set to run these tests")
    )

lusid_env = os.environ["LUSID_ENV"]
token = os.environ["FBN_ACCESS_TOKEN"]
host_vars = {}


@fixture
def setup_deployment():
    deployment_name = gen("configuration")
    print(f"\nRunning for deployment {deployment_name}...")
    yield SimpleNamespace(name=deployment_name)
    # Teardown: Clean up resources (if any) after the test
    print("\nTearing down resources...")
    fbnconfig.deployex(fbnconfig.Deployment(deployment_name, []), lusid_env, token)


def test_teardown(setup_deployment):
    # set it up
    fbnconfig.deployex(configuration.configure(setup_deployment), lusid_env, token)

    # Tear it down
    fbnconfig.deployex(fbnconfig.Deployment(setup_deployment.name, []), lusid_env, token)
    client = fbnconfig.create_client(lusid_env, token)
    try:
        client.request("get", "/configuration/api/sets/personal/robTest/rbCode")
    except HTTPStatusError as error:
        # check it was deleted
        assert error.response.status_code == 404


def test_create(setup_deployment):
    fbnconfig.deployex(configuration.configure(setup_deployment), lusid_env, token)
    client = fbnconfig.create_client(lusid_env, token)
    item = client.request(
        "get", f"/configuration/api/sets/personal/set1-{setup_deployment.name}/rbCode/items/username"
    )
    assert item.status_code == 200


def test_update(setup_deployment):
    fbnconfig.deployex(configuration.configure(setup_deployment), lusid_env, token)
    client = fbnconfig.create_client(lusid_env, token)
    item = client.request(
        "get", f"/configuration/api/sets/personal/set1-{setup_deployment.name}/rbCode/items/username"
    )

    assert item.status_code == 200


@pytest.mark.asyncio
async def test_item_list(setup_deployment):
    # GIVEN Item is deployed
    fbnconfig.deployex(configuration.configure(setup_deployment), lusid_env, token)
    type = "shared"
    scope = f"shared-{setup_deployment.name}"
    code = "testref2"
    key = "user"
    client = fbnconfig.create_client(lusid_env, token, is_async=True)
    # WHEN the fetch method is called for that resource
    response = await configuration.configuration.ItemResource.list(
        client=client,
        path_params={
            "setType": type,
            "setScope": scope,
            "setCode": code,
        }
    )
    # THEN the response is as expected
    expected_response = {
        "values": [
            {
                "keys": {
                    "setType": type,
                    "setScope": scope,
                    "setCode": code,
                    "key": key
                },
                "value": {
                    "id": f"item_{type}_{scope}_{code}_{key}".lower(),
                    "description": "something nice",
                    "key": key,
                    "value": "lucy-login",
                    "valueType": "text",
                    "isSecret": False,
                    "ref": f"config://{type}/{scope}/{code}/{key}",
                    "blockReveal": False,
                    "set": {
                        "$ref": f"set_{type}_{scope}_{code}".lower()
                    }
                }
            }
        ]
    }
    assert response == expected_response


@pytest.mark.asyncio
async def test_item_fetch(setup_deployment):
    # GIVEN Item is deployed
    fbnconfig.deployex(configuration.configure(setup_deployment), lusid_env, token)
    set_type = "shared"
    set_scope = f"shared-{setup_deployment.name}"
    set_code = "testref2"
    key = "user"
    client = fbnconfig.create_client(lusid_env, token, is_async=True)
    # WHEN the fetch method is called for that resource
    response = await configuration.configuration.ItemResource.fetch(
        client=client,
        keys={
            "setType": set_type,
            "setScope": set_scope,
            "setCode": set_code,
            "key": key
        }
    )
    # THEN the response is as expected
    expected_response = {
        "id": f"item_{set_type}_{set_scope}_{set_code}_{key}".lower(),
        "description": "Lucy test",
        "key": key,
        "value": "lucy-login",
        "valueType": "text",
        "isSecret": False,
        "ref": f"config://{set_type}/{set_scope}/{set_code}/{key}",
        "blockReveal": False,
        "set": {
            "$ref": f"set_{set_type}_{set_scope}_{set_code}".lower()
        }
    }
    assert response == expected_response


@pytest.mark.asyncio
async def test_set_list(setup_deployment):
    # GIVEN Set is deployed
    fbnconfig.deployex(configuration.configure(setup_deployment), lusid_env, token)
    client = fbnconfig.create_client(lusid_env, token, is_async=True)
    # WHEN the fetch method is called for that resource
    response = await configuration.configuration.SetResource.list(
        client=client,
        path_params={}
    )
    # THEN the response is as expected
    assert "values" in response and isinstance(response["values"], list)


@pytest.mark.asyncio
async def test_set_fetch(setup_deployment):
    # GIVEN Set is deployed
    fbnconfig.deployex(configuration.configure(setup_deployment), lusid_env, token)
    type = "shared"
    scope = f"shared-{setup_deployment.name}"
    code = "testref2"
    client = fbnconfig.create_client(lusid_env, token, is_async=True)
    # WHEN the fetch method is called for that resource
    response = await configuration.configuration.SetResource.fetch(
        client=client,
        keys={
            "type": type,
            "scope": scope,
            "code": code,
        }
    )
    # THEN the response is as expected
    expected_response = {
        "id": f"set_{type}_{scope}_{code}".lower(),
        "description": "something nice",
        "scope": scope,
        "code": code,
        "type": type,
    }
    assert response == expected_response



from fbnconfig import Deployment, configuration, scheduler

"""
An example configuration for defining configuration sets and their values.
The script configures the following entities:
- Configuration set
- Set item
- Image
- Job
- Schedule

The script configures an Image, Job and Schedule in order to show how you would
reference a configuration set item. For more information on these resources, see
`schedule.py` example.

More information can be found here:
https://support.lusid.com/knowledgebase/article/KA-01734/
https://support.lusid.com/knowledgebase/article/KA-01737/
https://support.lusid.com/knowledgebase/article/KA-01738/
"""


def configure(env):
    #
    # create a set and some items inside
    #
    cs = configuration.SetResource(
        id="set1",
        scope="sc1",
        code="cd1",
        type=configuration.SetType.PERSONAL,
        description="Example personal set resource",
    )

    username = configuration.ItemResource(
        id="user",
        set=cs,
        key="username",
        value="example-login",
        value_type=configuration.ValueType.TEXT,
        is_secret=False,
        description="Example key value pair representing a username",
    )

    passwd = configuration.ItemResource(
        id="password",
        set=cs,
        key="password",
        value="example-password",
        value_type=configuration.ValueType.TEXT,
        is_secret=False,
        description="Example key value pair representing a password",
    )

    #
    # add another set
    #
    set_ref = configuration.SetResource(
        id="set2",
        scope="sc2",
        code="cd1",
        type=configuration.SetType.SHARED,
        description="Example shared set resource",
    )

    ins_item = configuration.ItemResource(
        id="username2",
        set=set_ref,
        key="user",
        value="example-login-2",
        value_type=configuration.ValueType.TEXT,
        is_secret=False,
        description="Example key value pair representing a username",
    )

    #
    # create a scheduler job that uses the config for username and leaves password empty
    #
    image = scheduler.ImageResource(
        id="img1",
        source_image="docker.io/alpine:latest",
        dest_name="example-image-name",
        dest_tag="latest",
    )

    job = scheduler.JobResource(
        id="job1",
        scope="sc1",
        code="example-job-name",
        image=image,
        name="example-job",
        description="Example of a job using a configuration item",
        min_cpu="1",
        max_cpu="1",
        argument_definitions={
            "username": scheduler.EnvironmentArg(
                data_type="Configuration",
                required=False,
                description="A secret",
                order=1,
                default_value=username,
            ),
            "password": scheduler.EnvironmentArg(
                data_type="Configuration", required=False, description="A secret", order=1
            ),
        },
    )

    #
    # a schedule that uses config store for the password
    #
    schedule = scheduler.ScheduleResource(
        id="sch1",
        job=job,
        name="example-schedule",
        scope="sc1",
        code="cd1",
        expression="0 30 4 * * ? *",
        timezone="Europe/London",
        description="Example schedule",
        arguments={"username": username, "password": passwd},
    )

    return Deployment("configuration_example", [schedule, job, ins_item, cs, username, passwd])



import json
import string
from hashlib import sha256
from typing import Any, Dict

import httpx
from pydantic import BaseModel, Field, model_validator

from .resource_abc import CamelAlias, FetchMixin, ListMixin, Resource, register_resource
from .urlfmt import Urlfmt

u: string.Formatter = Urlfmt("/api/api/systemconfiguration/cutlabels")

_URL_FORMAT = "{base}/{code}"
GROUP = "Cut Label Definitions"


class CutTime(CamelAlias, BaseModel):
    hours: int
    minutes: int
    seconds: float


@register_resource(group=GROUP, keys=["code"])
class CutLabelResource(CamelAlias, BaseModel, Resource, ListMixin, FetchMixin):
    id: str = Field(init=True, exclude=True)
    code: str
    display_name: str
    description: str | None = None
    cut_local_time: CutTime
    time_zone: str

    @model_validator(mode="before")
    @classmethod
    def des_model(cls, data, info) -> Dict[str, Any]:
        if not isinstance(data, dict):
            return data
        # Handle id from context (for dump/undump)
        if info.context and info.context.get("id"):
            data = data | {"id": info.context.get("id")}
        return data

    def read(self, client: httpx.Client, old_state):
        url = u.format(_URL_FORMAT, code=old_state.code)
        return client.get(url).json()

    def create(self, client: httpx.Client):
        desired = self.model_dump(mode="json", exclude_none=True, by_alias=True)
        sorted_desired = json.dumps(desired, sort_keys=True)
        content_hash = sha256(sorted_desired.encode()).hexdigest()
        client.post(u.format("{base}"), json=desired)
        return {"code": self.code, "content_hash": content_hash}

    @staticmethod
    def delete(client: httpx.Client, old_state):
        code = old_state.code
        client.delete(u.format(_URL_FORMAT, code=code))

    def update(self, client: httpx.Client, old_state):
        if self.code != old_state.code:
            self.delete(client, old_state)
            return self.create(client)
        desired = self.model_dump(mode="json", exclude_none=True, by_alias=True)
        sorted_desired = json.dumps(desired, sort_keys=True)
        desired_hash = sha256(sorted_desired.encode()).hexdigest()
        if desired_hash == old_state.content_hash:
            return None
        client.put(u.format(_URL_FORMAT, code=self.code), json=desired)
        return {"code": self.code, "content_hash": desired_hash}

    @classmethod
    def create_resource_id(cls, data: dict) -> str:
        return f"cut-label_{data['code']}".lower()

    @classmethod
    def get_keys(cls, data: dict[str, Any]) -> dict[str, Any]:
        return {"code": data["code"]}

    @classmethod
    def from_api(cls, data: dict[str, Any]) -> dict[str, Any]:
        return {
            "id": cls.create_resource_id(data),
            "code": data["code"],
            "displayName": data["displayName"],
            "cutLocalTime": data["cutLocalTime"],
            "timeZone": data["timeZone"],
            "description": data.get("description")
        }

    @classmethod
    def list_endpoint(cls, **kwargs) -> str:
        return "/api/api/systemconfiguration/cutlabels"

    @classmethod
    def fetch_endpoint(cls) -> str:
        return "/api/api/systemconfiguration/cutlabels/{code}"

    def deps(self):
        return []


class CutLabelRef(CamelAlias, BaseModel):
    id: str = Field(init=True, exclude=True)
    code: str

    def attach(self, client: httpx.Client):
        try:
            url = u.format(_URL_FORMAT, code=self.code)
            response = client.get(url)
            return response.json()
        except httpx.HTTPStatusError as e:
            if e.response.status_code == 404:
                raise RuntimeError(f"CutLabel {self.code} not found")
            raise



import pathlib
import textwrap
from unittest import mock

import httpx
import pytest
from click.testing import CliRunner
from starlette.testclient import TestClient

from fbnconfig.http_client import response_hook_async
from server.app import app


@pytest.mark.respx(base_url="https://foo.lusid.com")
class DescribeServer:

    @pytest.fixture()
    def api(self):
        client = TestClient(app)
        client.event_hooks = {"response": [response_hook_async]}
        return client

    @pytest.fixture()
    def mock_resource(self):
        class MockResource:
            @classmethod
            async def fetch(cls, client, keys):
                return {"some": "data"}

        return MockResource

    def test_homepage(self, api):
        response = api.get("/api/fbnconfig/")
        assert response.status_code == 200
        assert response.json() == {"message": "Hello, Banana!", "status": "success"}

    def test_log(self, api, respx_mock):
        # given lusid has no deployments
        respx_mock.get("/api/api/customentities/~deployment").mock(
            return_value=httpx.Response(200, json={
                "values": []
            })
        )
        # when we request them
        headers = {
            "Authorization": "Bearer BINGOBANGO",
            "X-LUSID-Host": "https://foo.lusid.com",
        }
        response = api.get("/api/fbnconfig/log/something", headers=headers)
        assert response.status_code == 200
        request = respx_mock.calls.last.request
        assert request.method == "GET"
        assert request.headers["Authorization"] == "Bearer BINGOBANGO"
        assert request.headers["Host"] == "foo.lusid.com"

    def test_schema(self, api):
        # when we get the schema
        s = api.get("/api/fbnconfig/schema").json()
        # then it has defs and properties at the top level as expected
        assert s
        assert "$defs" in s
        assert "properties" in s

    def test_metadata(self, api):
        response = api.get("/api/fbnconfig/resources/metadata")
        assert response.status_code == 200
        assert isinstance(response.json(), list)

    def test_metadata_groups_containing_correct_resource_types(self, api):
        # If your resource is not coming back from this call you may need to add
        # your resource to __init__.py. Make sure you record your deployment.schema.json diff
        # with pytest_generate_tests as well when adding a new resource
        response = api.get("/api/fbnconfig/resources/metadata").json()
        grouped_resources = {
            group: {item["resource_type"] for item in response if item["group"] == group}
            for group in {item["group"] for item in response}
        }
        expected_grouped_resources = {
            "Identity & Access": {
                "PolicyResource",
                "PolicyCollectionResource",
                "RoleResource",
                "PolicyTemplateResource",
                "IdentityRoleResource",
                "UserResource",
                "RoleAssignment",
                "ApplicationResource"
            },
            "Compliance": {
                "ComplianceTemplateResource",
                "ComplianceRuleResource",
                "ReferenceListResource"
            },
            "Configuration Store": {
                "SetResource",
                "ItemResource",
                "SystemConfigResource"
            },
            "Portfolio Construction": {
                "CorporateActionSourceResource",
                "CustodianAccountResource",
                "TransactionTemplateResource",
                "TransactionTypeResource",
                "SideResource",
            },
            "Fund Accounting": {
                "CleardownModuleResource",
                "ChartOfAccountsResource",
                "AccountResource",
                "FundConfigurationResource",
                "GeneralLedgerResource",
                "PostingModuleResource",
            },
            "Data Model": {
                "DefinitionResource",
                "DefinitionOverrideResource",
                "IdentifierDefinitionResource",
                "DataTypeResource",
                "CustomDataModelResource",
                "EntityTypeResource",
                "EntityResource",
                "RelationalDatasetDefinitionResource",
            },
            "Cut Label Definitions": {
                "CutLabelResource",
            },
            "Drive": {
                "FolderResource",
                "FileResource",
            },
            "Horizon Integrations": {
                "IntegrationInstanceResource",
                "OptionalPropsResource",
            },
            "Luminesce": {
                "ViewResource",
                "InlinePropertiesResource",
            },
            "Notifications": {
                "SubscriptionResource",
                "NotificationResource",
            },
            "Valuation": {
                "RecipeResource",
            },
            "Scheduling": {
                "ImageResource",
                "JobResource",
                "ScheduleResource",
            },
            "Sequence": {
                "SequenceResource",
            },
            "Workflows": {
                "WorkerResource",
                "TaskDefinitionResource",
                "EventHandlerResource",
            },
            "Workspaces": {
                "WorkspaceResource",
                "WorkspaceItemResource",
            }
        }
        assert sorted(grouped_resources) == sorted(expected_grouped_resources)

    def test_metadata_resource_types_having_correct_parents(self, api):
        response = api.get("/api/fbnconfig/resources/metadata").json()
        resource_parent_map = {
            parent: {item["resource_type"] for item in response if item["parent"] == parent}
            for parent in {item["parent"] for item in response if item["parent"] is not None}
        }
        expected_resource_parent_map = {
            "ChartOfAccountsResource": {
                "AccountResource",
                "CleardownModuleResource",
                "GeneralLedgerResource",
                "PostingModuleResource",
            },
            "SetResource": {
                "ItemResource",
                "SystemConfigResource",
            },
            "EntityTypeResource": {
                "EntityResource",
            },
            "SubscriptionResource": {
                "NotificationResource",
            },
            "WorkspaceResource": {
                "WorkspaceItemResource",
            },
        }
        assert sorted(resource_parent_map) == sorted(expected_resource_parent_map)

    def test_metadata_resource_type_has_correct_schema(self, api):
        response = api.get("/api/fbnconfig/resources/metadata").json()
        policy_resource = next(
            item for item in response
            if item.get("resource_type") == "PolicyResource"
        )
        expected_policy_resource = {
            "resource_type": "PolicyResource",
            "group": "Identity & Access",
            "parent": None,
            "list_keys": [],
            "fetch_keys": ["scope", "code"],
            "filterable_columns": [],
            "supports_list": True,
            "supports_fetch": True,
            "supports_ref": True,
        }
        assert policy_resource == expected_policy_resource

    def test_metadata_resource_type_has_correct_list_keys(self, api):
        response = api.get("/api/fbnconfig/resources/metadata").json()
        workspace_item_resource = next(
            item for item in response
            if item.get("resource_type") == "WorkspaceItemResource"
        )
        expected_workspace_item_resource = {
            "resource_type": "WorkspaceItemResource",
            "group": "Workspaces",
            "parent": "WorkspaceResource",
            "list_keys": ["visibility", "name"],
            "fetch_keys": ["workspaceVisibility", "workspaceName", "group", "name"],
            "filterable_columns": [],
            "supports_list": True,
            "supports_fetch": True,
            "supports_ref": True,
        }
        assert workspace_item_resource == expected_workspace_item_resource

    @pytest.fixture()
    def example_dir(self):
        runner = CliRunner()
        with runner.isolated_filesystem():
            # given there is one example script in the examples folder
            script = textwrap.dedent("""\
                from fbnconfig import Deployment
                def configure(env):
                    return Deployment('test1', [])
            """)
            example_dir = pathlib.Path.cwd()
            with open("script.py", "w") as f:
                f.write(script)
            yield example_dir

    def test_examples(self, api, example_dir):
        # given a dir with one example script
        # and we mock the server to use it
        with mock.patch("server.app.get_examples_path", return_value=example_dir):
            # when we request examples
            examples = api.get("/api/fbnconfig/examples/").json()
            # there is one example
            assert examples == [{
                "name": "script",
                "path": "/api/fbnconfig/examples/script"
            }]

    def test_one_example(self, api, example_dir):
        # given a dir with one example script
        # and we mock the server to use it
        with mock.patch("server.app.get_examples_path", return_value=example_dir):
            # when we request an example that exists
            example = api.get("/api/fbnconfig/examples/script").json()
            # then we get a json with zero resources
            assert example == {
                "deploymentId": "test1",
                "resources": []
            }

    def test_fetch_success(self, api, mock_resource):
        # GIVEN a valid resource type
        resource_type = "SomeResource"
        # AND resource type exists in the registry
        resource_registry = {
            "SomeResource": {
                "class": mock_resource,
                "keys": ["scope", "code"]
            }
        }
        # WHEN the fetch endpoint is called for that resource type with valid keys
        with mock.patch(
            "server.api.RESOURCE_REGISTRY", resource_registry,
        ), mock.patch("server.api.fbnconfig.create_client"):
            response = api.post(
                f"/api/fbnconfig/resources/{resource_type}/fetch",
                json={
                    "keys": {
                        "scope": "some_scope",
                        "code": "some_code"
                    }
                }
            )
            print(response.json())
            # THEN the method returns 200 status code and the expected data
            assert response.status_code == 200
            assert response.json() == {"some": "data"}

    def test_fetch_error_invalid_resource_type(self, api):
        # GIVEN a resource type that doesn't match the expected camel case pattern
        resource_type = "invalid_resource"
        # WHEN the fetch endpoint is called for that resource type
        response = api.post(f"/api/fbnconfig/resources/{resource_type}/fetch")
        # THEN the method returns a 422 status code indicating validation error
        assert response.status_code == 422

    def test_fetch_error_no_such_resource(self, api):
        # GIVEN a valid resource type
        resource_type = "SomeResource"
        # AND resource type does not exist in the registry
        resource_registry = {}
        # WHEN the fetch endpoint is called for that resource type
        with mock.patch("server.api.RESOURCE_REGISTRY", resource_registry):
            response = api.post(f"/api/fbnconfig/resources/{resource_type}/fetch")
        # THEN the method returns a 404 status code indicating resource not found
        assert response.status_code == 404
        assert response.json() == {"detail": "Resource not found"}

    def test_fetch_error_fetch_method_not_implemented_on_resource(self, api):
        # GIVEN a valid resource type
        resource_type = "SomeResource"
        # AND resource type exists in the registry
        resource_registry = {
            "SomeResource": {
                # AND the resource class does not implement fetch method
                "class": mock.Mock,
                "keys": ["scope", "code"]
            }
        }
        # WHEN the fetch endpoint is called for that resource type
        with mock.patch(
            "server.api.RESOURCE_REGISTRY", resource_registry
        ):
            response = api.post(f"/api/fbnconfig/resources/{resource_type}/fetch")
            # THEN the method returns a 501 status code indicating not implemented
            assert response.status_code == 501
            assert response.json() == {"detail": "Method not implemented for this resource"}

    def test_fetch_error_missing_keys_param(self, api, mock_resource):
        # GIVEN a valid resource type
        resource_type = "SomeResource"
        # AND resource type exists in the registry
        resource_registry = {
            "SomeResource": {
                "class": mock_resource,
                "keys": ["scope", "code"]
            }
        }
        # WHEN the fetch endpoint is called for that resource type with missing required keys
        with mock.patch(
            "server.api.RESOURCE_REGISTRY", resource_registry,
        ), mock.patch("server.api.fbnconfig.create_client"):
            response = api.post(
                f"/api/fbnconfig/resources/{resource_type}/fetch",
                json={}
            )
            # THEN the method returns a 422 status code indicating validation error
            assert response.status_code == 422

    def test_fetch_error_missing_required_keys(self, api, mock_resource):
        # GIVEN a valid resource type
        resource_type = "SomeResource"
        # AND resource type exists in the registry
        resource_registry = {
            "SomeResource": {
                "class": mock_resource,
                "keys": ["scope", "code"]
            }
        }
        # WHEN the fetch endpoint is called for that resource type with missing required keys
        with mock.patch(
            "server.api.RESOURCE_REGISTRY", resource_registry,
        ), mock.patch("server.api.fbnconfig.create_client"):
            response = api.post(
                f"/api/fbnconfig/resources/{resource_type}/fetch",
                json={
                    "keys": {}
                }
            )
            # THEN the method returns a 422 status code indicating validation error
            assert response.status_code == 422

    def test_fetch_error_wrong_required_keys(self, api, mock_resource):
        # GIVEN a valid resource type
        resource_type = "SomeResource"
        # AND resource type exists in the registry
        resource_registry = {
            "SomeResource": {
                "class": mock_resource,
                "keys": ["scope", "code"]
            }
        }
        # WHEN the fetch endpoint is called for that resource type with wrong required keys
        with mock.patch(
            "server.api.RESOURCE_REGISTRY", resource_registry,
        ), mock.patch("server.api.fbnconfig.create_client"):
            response = api.post(
                f"/api/fbnconfig/resources/{resource_type}/fetch",
                json={
                    "wrong_key": "some_value"
                }
            )
            # THEN the method returns a 422 status code indicating validation error
            assert response.status_code == 422



import os
from types import SimpleNamespace
from unittest.mock import ANY

import pytest
from httpx import HTTPStatusError

import fbnconfig
import tests.integration.schedule as schedule
from tests.integration.generate_test_name import gen

if os.getenv("FBN_ACCESS_TOKEN") is None or os.getenv("LUSID_ENV") is None:
    raise (
        RuntimeError("Both FBN_ACCESS_TOKEN and LUSID_ENV variables need to be set to run these tests")
    )

lusid_env = os.environ["LUSID_ENV"]
token = os.environ["FBN_ACCESS_TOKEN"]
host_vars = {}
client = fbnconfig.create_client(lusid_env, token)


@pytest.fixture(scope="module")
def setup_deployment():
    deployment_name = gen("schedule")
    print(f"\nRunning for deployment {deployment_name}...")
    yield SimpleNamespace(name=deployment_name)
    # Teardown: Clean up resources (if any) after the test
    print("\nTearing down resources...")
    fbnconfig.deployex(fbnconfig.Deployment(deployment_name, []), lusid_env, token)


def test_teardown(setup_deployment):
    # create first
    fbnconfig.deployex(schedule.configure(setup_deployment), lusid_env, token)
    fbnconfig.deployex(fbnconfig.Deployment(setup_deployment.name, []), lusid_env, token)
    try:
        client.get(f"/scheduler2/api/schedules/{setup_deployment.name}/python-sched")
    except HTTPStatusError as error:
        assert error.response.status_code == 404


def test_create(setup_deployment):
    fbnconfig.deployex(schedule.configure(setup_deployment), lusid_env, token)
    search = client.get(f"/scheduler2/api/schedules/{setup_deployment.name}/python-sched")
    assert search.status_code == 200


def test_update(setup_deployment):
    fbnconfig.deployex(schedule.configure(setup_deployment), lusid_env, token)
    fbnconfig.deployex(schedule.configure(setup_deployment), lusid_env, token)
    search = client.get(f"/scheduler2/api/schedules/{setup_deployment.name}/python-sched")
    assert search.status_code == 200


@pytest.mark.asyncio
async def test_schedule_list_with_filter(setup_deployment):
    # GIVEN Schedule is deployed
    fbnconfig.deployex(schedule.configure(setup_deployment), lusid_env, token)
    scope = setup_deployment.name
    client = fbnconfig.create_client(lusid_env, token, is_async=True)
    # WHEN calling list with filter for the deployment name
    result = await schedule.scheduler.ScheduleResource.list(
        client=client,
        query_params={"filter": f"scheduleIdentifier.scope eq '{scope}'"},
    )
    # THEN the result has correct structure
    expected_result = {
        "values": [
            {
                "keys": {
                    "scope": scope,
                    "code": "python-sched"
                },
                "value": {
                    "id": f"schedule_{scope}_python-sched".lower(),
                    "scope": scope,
                    "code": "python-sched",
                    "arguments": {
                        "arg1": "mynotdefaultvalue"
                    },
                    "author": "",
                    "description": "my awesome schedule",
                    "enabled": True,
                    "expression": "0 30 4 * * ? *",
                    "name": "my-schedule",
                    "owner": "",
                    "timezone": "Europe/London",
                    "useAsAuth": {
                        "$ref": ANY
                    },
                    "job": {
                        "$ref": f"job_{scope}_job".lower()
                    }
                }
            }
        ]
    }
    assert result["values"] == expected_result["values"]


@pytest.mark.asyncio
async def test_schedule_fetch(setup_deployment):
    # GIVEN Schedule is deployed
    fbnconfig.deployex(schedule.configure(setup_deployment), lusid_env, token)
    scope = setup_deployment.name
    code = "python-sched"
    client = fbnconfig.create_client(lusid_env, token, is_async=True)
    # WHEN the fetch method is called for that resource
    response = await schedule.scheduler.ScheduleResource.fetch(
        client=client,
        keys={
            "scope": scope,
            "code": code,
        }
    )
    # THEN the response is as expected
    expected_response = {
        "id": f"schedule_{scope}_python-sched".lower(),
        "scope": scope,
        "code": "python-sched",
        "arguments": {
            "arg1": "mynotdefaultvalue"
        },
        "author": "",
        "description": "my awesome schedule",
        "enabled": True,
        "expression": "0 30 4 * * ? *",
        "name": "my-schedule",
        "owner": "",
        "timezone": "Europe/London",
        "useAsAuth": {
            "$ref": ANY
        },
        "job": {
            "$ref": f"job_{scope}_job".lower()
        }
    }
    assert response == expected_response


@pytest.mark.asyncio
async def test_image_fetch(setup_deployment):
    # GIVEN Image is deployed
    fbnconfig.deployex(schedule.configure(setup_deployment), lusid_env, token)
    dest_name = "beany"
    dest_tag = "v6.0"
    client = fbnconfig.create_client(lusid_env, token, is_async=True)
    # WHEN the fetch method is called for that resource
    response = await schedule.scheduler.ImageResource.fetch(
        client=client,
        keys={
            "destName": dest_name,
            "destTag": dest_tag,
        }
    )
    # THEN the response is as expected
    expected_response = {
        "id": f"image_{dest_name}_{dest_tag}".lower(),
        "destName": dest_name,
        "destTag": dest_tag,
        "sourceImage": f"{dest_name}:{dest_tag}"
    }
    assert response == expected_response


@pytest.mark.asyncio
async def test_job_list_with_filter(setup_deployment):
    # GIVEN Job is deployed
    fbnconfig.deployex(schedule.configure(setup_deployment), lusid_env, token)
    scope = setup_deployment.name
    code = "job"
    client = fbnconfig.create_client(lusid_env, token, is_async=True)
    # WHEN calling list with filter for the deployment name
    result = await schedule.scheduler.JobResource.list(
        client=client,
        query_params={"filter": f"jobId.scope eq '{scope}' and jobId.code eq '{code}'"},
    )
    # THEN the result has correct structure
    expected_result = {
        "values": [
            {
                "keys": {
                    "scope": scope,
                    "code": code
                },
                "value": {
                    "id": f"job_{scope}_{code}".lower(),
                    "scope": scope,
                    "code": code,
                    "commandLineArgumentSeparator": " ",
                    "dateCreated": ANY,
                    "description": "something nice",
                    "author": "",
                    "maxCpu": "2",
                    "maxMemory": "200Mi",
                    "minCpu": "1",
                    "minMemory": "100Mi",
                    "ttl": 300,
                    "image": {
                        "$ref": "image_beany_v6.0".lower()
                    },
                    "name": "RobsJob",
                    "argumentDefinitions": {
                        "arg1": {
                            "constraints": "None",
                            "dataType": "String",
                            "defaultValue": "mydefaultvalue",
                            "description": "My argument",
                            "order": 1,
                            "passedAs": "EnvironmentVariable",
                            "required": False
                        }
                    },
                }
            }
        ]
    }
    assert result["values"] == expected_result["values"]



from __future__ import annotations

from typing import Any, Dict

import httpx
from pydantic import BaseModel, Field, model_validator

from fbnconfig.resource_abc import CamelAlias, FetchMixin, ListMixin, Resource, register_resource

GROUP = "Sequence"


@register_resource(group=GROUP, keys=["scope", "code"])
class SequenceResource(CamelAlias, BaseModel, Resource, ListMixin, FetchMixin):
    resource_type: str = Field(default="SequenceResource", exclude=True)
    id: str = Field(exclude=True)
    scope: str
    code: str
    increment: int | None = None
    min_value: int | None = None
    max_value: int | None = None
    start: int | None = None
    cycle: bool | None = None
    pattern: str | None = None

    @model_validator(mode="before")
    @classmethod
    def des_model(cls, data: Dict[str, Any], info) -> Dict[str, Any]:
        if isinstance(data.get("id"), dict):
            # promote scope and code
            data = data | data.pop("id")
        if info.context and info.context.get("id"):
            return data | {"id": info.context.get("id")}
        return data

    def read(self, client, old_state) -> Dict[str, Any]:
        return client.request("get", f"/api/api/sequences/{self.scope}/{self.code}").json()

    def create(self, client: httpx.Client):
        desired = self.model_dump(mode="json", exclude_none=True, by_alias=True,
            exclude={"scope"})
        client.request("POST", f"/api/api/sequences/{self.scope}", json=desired)
        return {"scope": self.scope, "code": self.code}

    def update(self, client: httpx.Client, old_state):
        if [old_state.scope, old_state.code] != [self.scope, self.code]:
            raise (RuntimeError("Cannot change the scope/code on a sequence"))
        remote = self.read(client, old_state)
        desired = self.model_dump(
            mode="json", exclude_none=True, exclude={"scope", "code"}, by_alias=True
        )
        effective = remote | desired
        if effective == remote:
            return None
        raise RuntimeError("Cannot modify a sequence")

    @staticmethod
    def delete(client, old_state):
        raise RuntimeError("Cannot delete a sequence")

    @classmethod
    def create_resource_id(cls, data: dict) -> str:
        return f"sequence_{data['scope']}_{data['code']}".lower()

    @classmethod
    def from_api(cls, data: dict[str, Any]) -> dict[str, Any]:
        data.pop("links", None)
        data.pop("value", None)
        data = data | data["id"]
        data["id"] = cls.create_resource_id(data)
        return data

    @classmethod
    def get_keys(cls, data: dict[str, Any]) -> dict[str, Any]:
        return data["id"]

    @classmethod
    def fetch_endpoint(cls) -> str:
        return "/api/api/sequences/{scope}/{code}"

    @classmethod
    def list_endpoint(cls, **kwargs) -> str:
        return "/api/api/sequences"

    def deps(self):
        return []



import json
from pathlib import Path

import pytest


def load_data(path: str) -> dict:
    return json.loads(Path(path).read_text())


@pytest.fixture
def testdata_loader():
    """Returns a function to load test data for any resource."""
    def _load(resource_name: str, file_type: str) -> dict:
        return load_data(f"tests/unit/testdata/{resource_name}/{file_type}.json")
    return _load


__all__ = ["testdata_loader"]



import os
from types import SimpleNamespace

import pytest
from pytest import fixture

import fbnconfig
import tests.integration.sequence as sequence
from tests.integration.generate_test_name import gen

if os.getenv("FBN_ACCESS_TOKEN") is None or os.getenv("LUSID_ENV") is None:
    raise (
        RuntimeError("Both FBN_ACCESS_TOKEN and LUSID_ENV variables need to be set to run these tests")
    )

lusid_env = os.environ["LUSID_ENV"]
token = os.environ["FBN_ACCESS_TOKEN"]
host_vars = {}
client = fbnconfig.create_client(lusid_env, token)


@fixture(scope="module")
def setup_deployment():
    deployment_name = gen("sequence")
    print(f"\nRunning for deployment {deployment_name}...")
    yield SimpleNamespace(name=deployment_name)
    # Cant tear down sequence resources as no api to delete


def test_create(setup_deployment):
    fbnconfig.deployex(sequence.configure(setup_deployment), lusid_env, token)
    client.get(f"/api/api/sequences/{setup_deployment.name}/seq1").json()


def test_nochange(setup_deployment):
    fbnconfig.deployex(sequence.configure(setup_deployment), lusid_env, token)
    update = fbnconfig.deployex(sequence.configure(setup_deployment), lusid_env, token)
    assert [a.change for a in update if a.type == "SequenceResource"] == ["nochange"]


@pytest.mark.asyncio
async def test_sequence_list_with_filter(setup_deployment):
    # GIVEN Sequence is deployed
    fbnconfig.deployex(sequence.configure(setup_deployment), lusid_env, token)
    scope = setup_deployment.name
    code = "seq1"
    client = fbnconfig.create_client(lusid_env, token, is_async=True)
    # WHEN calling list with filter for the deployment name
    result = await sequence.sequence.SequenceResource.list(
        client=client,
        query_params={"filter": f"id.code eq '{code}' and id.scope eq '{scope}'"},
        path_params={}
    )
    # THEN the result has correct structure
    expected_result = {
        "values": [
            {
                "keys": {
                    "code": code,
                    "scope": scope
                },
                "value": {
                    "id": f"sequence_{scope}_{code}".lower(),
                    "code": code,
                    "scope": scope,
                    "cycle": False,
                    "increment": 1,
                    "start": 1,
                    "minValue": 1,
                    "maxValue": 9223372036854775807,
                }
            }
        ]
    }
    assert result == expected_result


@pytest.mark.asyncio
async def test_sequence_fetch(setup_deployment):
    # GIVEN Sequence is deployed
    fbnconfig.deployex(sequence.configure(setup_deployment), lusid_env, token)
    scope = setup_deployment.name
    code = "seq1"
    client = fbnconfig.create_client(lusid_env, token, is_async=True)
    # WHEN the fetch method is called for that resource
    response = await sequence.sequence.SequenceResource.fetch(
        client=client,
        keys={
            "code": code,
            "scope": scope
        }
    )
    # THEN the response is as expected
    expected_response = {
        "id": f"sequence_{scope}_{code}".lower(),
        "code": code,
        "scope": scope,
        "cycle": False,
        "increment": 1,
        "start": 1,
        "minValue": 1,
        "maxValue": 9223372036854775807,
    }
    assert response == expected_response



import os
from types import SimpleNamespace

import pytest
from httpx import HTTPStatusError
from pytest import fixture

import fbnconfig
from fbnconfig import Deployment, fund_accounting, posting_module, property
from tests.integration.generate_test_name import gen

if os.getenv("FBN_ACCESS_TOKEN") is None or os.getenv("LUSID_ENV") is None:
    raise (
        RuntimeError("Both FBN_ACCESS_TOKEN and LUSID_ENV variables need to be set to run these tests")
    )

lusid_env = os.environ["LUSID_ENV"]
token = os.environ["FBN_ACCESS_TOKEN"]
host_vars = {}


def configure(env):
    deployment_name = getattr(env, "name", "chart_of_accounts")

    perp_prop = fund_accounting.PropertyValue(
        property_key=property.DefinitionRef(
            id="one", domain=property.Domain.ChartOfAccounts, scope="sc1", code="cd1"
        ),
        label_value="Hello"
    )

    chart_of_account = fund_accounting.ChartOfAccountsResource(
        id="example_id",
        scope=f"testScope-{deployment_name}",
        code="testCode",
        display_name="example_display_name",
        description="example_description",
        properties=[perp_prop]
    )

    perp_prop_two = fund_accounting.PropertyValue(
        property_key=property.DefinitionRef(
            id="two", domain=property.Domain.Account, scope="sc2", code="cd2"
        ),
        label_value="Goodbye"
    )

    account = fund_accounting.AccountResource(
        id=f"example_id-{deployment_name}",
        chart_of_accounts=chart_of_account,
        code=f"account_code-{deployment_name}",
        description="example_desc",
        type=fund_accounting.AccountType.ASSET,
        status=fund_accounting.AccountStatus.ACTIVE,
        control="Manual",
        properties=[perp_prop_two]
    )

    rule = posting_module.PostingModuleRule(
        rule_id="rule_id",
        general_ledger_account_code=account,
        rule_filter="SourceType eq 'LusidTransaction'"
    )

    post_mod = posting_module.PostingModuleResource(
        id="posting_module_id",
        chart_of_accounts=chart_of_account,
        code=f"module_code-{deployment_name}",
        display_name="example_display_name",
        description="example_description",
        rules=[rule]
    )

    return Deployment(deployment_name, [account, chart_of_account, post_mod])


def create_new_properties(client):
    print("Creating new property")

    delete_properties(client)

    # Creates new property to add to the chart of accounts
    chart_property = {
        "domain": "ChartOfAccounts",
        "scope": "sc1",
        "code": "cd1",
        "valueRequired": False,
        "displayName": "My Property Display Name",
        "dataTypeId": {
            "scope": "system",
            "code": "string"
        },
        "lifeTime": "Perpetual",
    }
    client.post("/api/api/propertydefinitions", json=chart_property)

    # Creates new property add to the chart of accounts
    account_property = {
        "domain": "Account",
        "scope": "sc2",
        "code": "cd2",
        "valueRequired": False,
        "displayName": "My Property Display Name",
        "dataTypeId": {
            "scope": "system",
            "code": "string"
        },
        "lifeTime": "Perpetual",
    }
    client.post("/api/api/propertydefinitions", json=account_property)


def delete_properties(client):
    print("Deleting properties")

    # Deletes test property for chart of account if it exsits
    try:
        client.delete("/api/api/propertydefinitions/ChartOfAccounts/sc1/cd1")
    except Exception:
        pass
    # Deletes test property for account if it exsits
    try:
        client.delete("/api/api/propertydefinitions/Account/sc2/cd2")
    except Exception:
        pass


@fixture
def setup_deployment():
    client = fbnconfig.create_client(lusid_env, token)
    create_new_properties(client)

    deployment_name = gen("chart_of_accounts")
    print(f"\nRunning for deployment {deployment_name}...")

    yield SimpleNamespace(name=deployment_name)
    # Teardown: Clean up resources (if any) after the test
    print("\nTearing down resources...")
    fbnconfig.deployex(fbnconfig.Deployment(deployment_name, []), lusid_env, token)
    delete_properties(client)


def test_teardown(setup_deployment):
    deployment_name = setup_deployment.name
    client = fbnconfig.create_client(lusid_env, token)
    # setup deployment
    fbnconfig.deployex(configure(setup_deployment), lusid_env, token)

    chart_url = f"/api/api/chartofaccounts/testScope-{deployment_name}/testCode"
    account_url = chart_url + f"/accounts/account_code-{deployment_name}"
    posting_mod_url = chart_url + f"/postingmodules/module_code-{deployment_name}"

    # check it exists
    client.request("get", chart_url)
    client.request("get", account_url)
    client.request("get", posting_mod_url)
    # Tear it down
    fbnconfig.deployex(fbnconfig.Deployment(deployment_name, []), lusid_env, token)

    # check account was deleted
    try:
        client.request("get", account_url)
    except HTTPStatusError as error:
        assert error.response.status_code == 404
    # check chart was deleted
    try:
        client.request("get", posting_mod_url)
    except HTTPStatusError as error:
        assert error.response.status_code == 404
    # check chart was deleted
    try:
        client.request("get", chart_url)
    except HTTPStatusError as error:
        assert error.response.status_code == 404


def test_create(setup_deployment):
    deployment_name = setup_deployment.name
    fbnconfig.deployex(configure(setup_deployment), lusid_env, token)
    client = fbnconfig.create_client(lusid_env, token)
    chart_url = f"/api/api/chartofaccounts/testScope-{deployment_name}/testCode"
    search_chart = client.request("get", chart_url)
    account_url = chart_url + f"/accounts/account_code-{deployment_name}"
    search_account = client.request("get", account_url)
    assert search_chart.status_code == 200
    assert search_account.status_code == 200


def test_update_nochange(setup_deployment):
    fbnconfig.deployex(configure(setup_deployment), lusid_env, token)
    update = fbnconfig.deployex(configure(setup_deployment), lusid_env, token)
    assert [a.change for a in update if a.type == "ChartOfAccountsResource"] == ["nochange"]
    assert [a.change for a in update if a.type == "AccountResource"] == ["nochange"]
    assert [a.change for a in update if a.type == "PostingModuleResource"] == ["nochange"]


@pytest.mark.asyncio
async def test_chart_of_accounts_list_with_filter(setup_deployment):
    # GIVEN ChartOfAccounts is deployed
    fbnconfig.deployex(configure(setup_deployment), lusid_env, token)
    deployment_name = setup_deployment.name
    code = "testCode"
    scope = f"testScope-{deployment_name}"
    client = fbnconfig.create_client(lusid_env, token, is_async=True)
    # WHEN calling list with filter for the deployment name
    result = await fund_accounting.ChartOfAccountsResource.list(
        client=client,
        query_params={"filter": f"id.code eq '{code}' and id.scope eq '{scope}'"},
        path_params={}
    )
    # THEN the result has correct structure
    expected_result = {
        "values": [
            {
                "keys": {
                    "code": code,
                    "scope": scope
                },
                "value": {
                    "id": f"chart-of-account_{scope}_{code}".lower(),
                    "code": code,
                    "scope": scope,
                    "description": "example_description",
                    "displayName": "example_display_name",
                    "properties": {}
                }
            }
        ]
    }
    assert result["values"] == expected_result["values"]


@pytest.mark.asyncio
async def test_chart_of_accounts_fetch(setup_deployment):
    # GIVEN ChartOfAccounts is deployed
    fbnconfig.deployex(configure(setup_deployment), lusid_env, token)
    deployment_name = setup_deployment.name
    code = "testCode"
    scope = f"testScope-{deployment_name}"
    client = fbnconfig.create_client(lusid_env, token, is_async=True)
    # WHEN the fetch method is called for that resource
    response = await fund_accounting.ChartOfAccountsResource.fetch(
        client=client,
        keys={
            "code": code,
            "scope": scope
        }
    )
    # THEN the response is as expected
    expected_response = {
        "id": f"chart-of-account_{scope}_{code}".lower(),
        "code": code,
        "scope": scope,
        "description": "example_description",
        "displayName": "example_display_name",
        "properties": {}
    }
    assert response == expected_response


@pytest.mark.asyncio
async def test_account_list_with_filter(setup_deployment):
    # GIVEN Account is deployed
    fbnconfig.deployex(configure(setup_deployment), lusid_env, token)
    deployment_name = setup_deployment.name
    account_code = f"account_code-{deployment_name}"
    coa_code = "testCode"
    coa_scope = f"testScope-{deployment_name}"
    client = fbnconfig.create_client(lusid_env, token, is_async=True)
    # WHEN calling list with filter for the deployment name
    result = await fund_accounting.AccountResource.list(
        client=client,
        query_params={"filter": f"code eq '{account_code}'"},
        path_params={
            "chartOfAccountsScope": coa_scope,
            "chartOfAccountsCode": coa_code,
        }
    )
    # THEN the result has correct structure
    expected_result = {
        "values": [
            {
                "keys": {
                    "chartOfAccountsScope": coa_scope,
                    "chartOfAccountsCode": coa_code,
                    "code": account_code
                },
                "value": {
                    "id": f"account_{coa_scope}_{coa_code}_{account_code}".lower(),
                    "chartOfAccounts": {
                        "$ref": f"chart-of-account_{coa_scope}_{coa_code}".lower()
                    },
                    "code": account_code,
                    "description": "example_desc",
                    "type": "Asset",
                    "status": "Active",
                    "control": "Manual",
                    "properties": {}
                }
            }
        ]
    }
    assert result["values"] == expected_result["values"]


@pytest.mark.asyncio
async def test_account_fetch(setup_deployment):
    # GIVEN Account is deployed
    fbnconfig.deployex(configure(setup_deployment), lusid_env, token)
    deployment_name = setup_deployment.name
    account_code = f"account_code-{deployment_name}"
    coa_code = "testCode"
    coa_scope = f"testScope-{deployment_name}"
    client = fbnconfig.create_client(lusid_env, token, is_async=True)
    # WHEN the fetch method is called for that resource
    response = await fund_accounting.AccountResource.fetch(
        client=client,
        keys={
            "chartOfAccountsScope": coa_scope,
            "chartOfAccountsCode": coa_code,
            "code": account_code
        }
    )
    # THEN the response is as expected
    expected_response = {
        "id": f"account_{coa_scope}_{coa_code}_{account_code}".lower(),
        "chartOfAccounts": {
            "$ref": f"chart-of-account_{coa_scope}_{coa_code}".lower()
        },
        "code": account_code,
        "description": "example_desc",
        "type": "Asset",
        "status": "Active",
        "control": "Manual",
        "properties": {}
    }
    assert response == expected_response


@pytest.mark.asyncio
async def test_posting_module_list_with_filter(setup_deployment):
    # GIVEN PostingModule is deployed
    fbnconfig.deployex(configure(setup_deployment), lusid_env, token)
    deployment_name = setup_deployment.name
    posting_code = f"module_code-{deployment_name}"
    account_code = f"account_code-{deployment_name}"
    coa_code = "testCode"
    coa_scope = f"testScope-{deployment_name}"
    client = fbnconfig.create_client(lusid_env, token, is_async=True)
    # WHEN calling list with filter for the deployment name
    result = await posting_module.PostingModuleResource.list(
        client=client,
        query_params={"filter": f"postingModuleCode eq '{posting_code}'"},
        path_params={
            "chartOfAccountsScope": coa_scope,
            "chartOfAccountsCode": coa_code,
        }
    )
    # THEN the result has correct structure
    expected_result = {
        "values": [
            {
                "keys": {
                    "chartOfAccountsScope": coa_scope,
                    "chartOfAccountsCode": coa_code,
                    "code": posting_code
                },
                "value": {
                    "id": f"posting-module_{coa_scope}_{coa_code}_{posting_code}".lower(),
                    "chartOfAccounts": {
                        "$ref": f"chart-of-account_{coa_scope}_{coa_code}".lower()
                    },
                    "code": posting_code,
                    "displayName": "example_display_name",
                    "description": "example_description",
                    "rules": [
                        {
                            "ruleId": "rule_id",
                            "generalLedgerAccountCode": {
                                "$ref": f"account_{coa_scope}_{coa_code}_{account_code}".lower()
                            },
                            "ruleFilter": "SourceType eq 'LusidTransaction'",
                        }
                    ]
                }
            }
        ]
    }
    assert result["values"] == expected_result["values"]


@pytest.mark.asyncio
async def test_posting_module_fetch(setup_deployment):
    # GIVEN PostingModule is deployed
    fbnconfig.deployex(configure(setup_deployment), lusid_env, token)
    deployment_name = setup_deployment.name
    posting_code = f"module_code-{deployment_name}"
    account_code = f"account_code-{deployment_name}"
    coa_code = "testCode"
    coa_scope = f"testScope-{deployment_name}"
    client = fbnconfig.create_client(lusid_env, token, is_async=True)
    # WHEN the fetch method is called for that resource
    response = await posting_module.PostingModuleResource.fetch(
        client=client,
        keys={
            "chartOfAccountsScope": coa_scope,
            "chartOfAccountsCode": coa_code,
            "code": posting_code
        }
    )
    # THEN the response is as expected
    expected_response = {
        "id": f"posting-module_{coa_scope}_{coa_code}_{posting_code}".lower(),
        "chartOfAccounts": {
            "$ref": f"chart-of-account_{coa_scope}_{coa_code}".lower()
        },
        "code": posting_code,
        "displayName": "example_display_name",
        "description": "example_description",
        "rules": [
            {
                "ruleId": "rule_id",
                "generalLedgerAccountCode": {
                    "$ref": f"account_{coa_scope}_{coa_code}_{account_code}".lower()
                },
                "ruleFilter": "SourceType eq 'LusidTransaction'",
            }
        ]
    }
    assert response == expected_response
